@@ -1,18 +1,18 @@
+/* SPDX-License-Identifier: GPL-2.0 */
 /*
- * Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
- *
- * See doc/protocol.md for more info
+ * Copyright (C) 2015-2018 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
  */
 
-#ifndef MESSAGES_H
-#define MESSAGES_H
+#ifndef _WG_MESSAGES_H
+#define _WG_MESSAGES_H
 
-#include "crypto/curve25519.h"
-#include "crypto/chacha20poly1305.h"
-#include "crypto/blake2s.h"
+#include <zinc/curve25519.h>
+#include <zinc/chacha20poly1305.h>
+#include <zinc/blake2s.h>
 
 #include <linux/kernel.h>
 #include <linux/param.h>
+#include <linux/skbuff.h>
 
 enum noise_lengths {
 	NOISE_PUBLIC_KEY_LEN = CURVE25519_POINT_SIZE,
@@ -25,8 +25,8 @@ enum noise_lengths {
 #define noise_encrypted_len(plain_len) (plain_len + NOISE_AUTHTAG_LEN)
 
 enum cookie_values {
-	COOKIE_SECRET_MAX_AGE = 2 * 60 * HZ,
-	COOKIE_SECRET_LATENCY = 5 * HZ,
+	COOKIE_SECRET_MAX_AGE = 2 * 60,
+	COOKIE_SECRET_LATENCY = 5,
 	COOKIE_NONCE_LEN = XCHACHA20POLY1305_NONCELEN,
 	COOKIE_LEN = 16
 };
@@ -40,17 +40,17 @@ enum counter_values {
 enum limits {
 	REKEY_AFTER_MESSAGES = U64_MAX - 0xffff,
 	REJECT_AFTER_MESSAGES = U64_MAX - COUNTER_WINDOW_SIZE - 1,
-	REKEY_TIMEOUT = 5 * HZ,
-	REKEY_TIMEOUT_JITTER_MAX = HZ / 3,
-	REKEY_AFTER_TIME = 120 * HZ,
-	REJECT_AFTER_TIME = 180 * HZ,
-	INITIATIONS_PER_SECOND = HZ / 50,
-	MAX_PEERS_PER_DEVICE = U16_MAX,
-	KEEPALIVE_TIMEOUT = 10 * HZ,
-	MAX_TIMER_HANDSHAKES = (90 * HZ) / REKEY_TIMEOUT,
-	MAX_QUEUED_INCOMING_HANDSHAKES = 4096,
-	MAX_BURST_INCOMING_HANDSHAKES = 16,
-	MAX_QUEUED_OUTGOING_PACKETS = 1024
+	REKEY_TIMEOUT = 5,
+	REKEY_TIMEOUT_JITTER_MAX_JIFFIES = HZ / 3,
+	REKEY_AFTER_TIME = 120,
+	REJECT_AFTER_TIME = 180,
+	INITIATIONS_PER_SECOND = 50,
+	MAX_PEERS_PER_DEVICE = 1U << 20,
+	KEEPALIVE_TIMEOUT = 10,
+	MAX_TIMER_HANDSHAKES = 90 / REKEY_TIMEOUT,
+	MAX_QUEUED_INCOMING_HANDSHAKES = 4096, /* TODO: replace this with DQL */
+	MAX_STAGED_PACKETS = 128,
+	MAX_QUEUED_PACKETS = 1024 /* TODO: replace this with DQL */
 };
 
 enum message_type {
@@ -58,8 +58,7 @@ enum message_type {
 	MESSAGE_HANDSHAKE_INITIATION = 1,
 	MESSAGE_HANDSHAKE_RESPONSE = 2,
 	MESSAGE_HANDSHAKE_COOKIE = 3,
-	MESSAGE_DATA = 4,
-	MESSAGE_TOTAL = 5
+	MESSAGE_DATA = 4
 };
 
 struct message_header {
@@ -110,34 +109,20 @@ struct message_data {
 	u8 encrypted_data[];
 };
 
-#define message_data_len(plain_len) (noise_encrypted_len(plain_len) + sizeof(struct message_data))
+#define message_data_len(plain_len)                                            \
+	(noise_encrypted_len(plain_len) + sizeof(struct message_data))
 
 enum message_alignments {
 	MESSAGE_PADDING_MULTIPLE = 16,
 	MESSAGE_MINIMUM_LENGTH = message_data_len(0)
 };
 
-#define SKB_HEADER_LEN (max(sizeof(struct iphdr), sizeof(struct ipv6hdr)) + sizeof(struct udphdr) + NET_SKB_PAD)
-#define DATA_PACKET_HEAD_ROOM ALIGN(sizeof(struct message_data) + SKB_HEADER_LEN, 4)
+#define SKB_HEADER_LEN                                                         \
+	(max(sizeof(struct iphdr), sizeof(struct ipv6hdr)) +                   \
+	 sizeof(struct udphdr) + NET_SKB_PAD)
+#define DATA_PACKET_HEAD_ROOM                                                  \
+	ALIGN(sizeof(struct message_data) + SKB_HEADER_LEN, 4)
 
-enum {
-	HANDSHAKE_DSCP = 0b10001000 /* AF41, plus 00 ECN */
-};
+enum { HANDSHAKE_DSCP = 0x88 /* AF41, plus 00 ECN */ };
 
-static inline enum message_type message_determine_type(void *src, size_t src_len)
-{
-	struct message_header *header = src;
-	if (unlikely(src_len < sizeof(struct message_header)))
-		return MESSAGE_INVALID;
-	if (header->type == cpu_to_le32(MESSAGE_DATA) && src_len >= MESSAGE_MINIMUM_LENGTH)
-		return MESSAGE_DATA;
-	if (header->type == cpu_to_le32(MESSAGE_HANDSHAKE_INITIATION) && src_len == sizeof(struct message_handshake_initiation))
-		return MESSAGE_HANDSHAKE_INITIATION;
-	if (header->type == cpu_to_le32(MESSAGE_HANDSHAKE_RESPONSE) && src_len == sizeof(struct message_handshake_response))
-		return MESSAGE_HANDSHAKE_RESPONSE;
-	if (header->type == cpu_to_le32(MESSAGE_HANDSHAKE_COOKIE) && src_len == sizeof(struct message_handshake_cookie))
-		return MESSAGE_HANDSHAKE_COOKIE;
-	return MESSAGE_INVALID;
-}
-
-#endif
+#endif /* _WG_MESSAGES_H */
